package model;

import org.json.JSONArray;
import org.json.JSONObject;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public class GameState
{
	long[] fish; //Array of three bitboards: one for every fish value
	Player current_player;
	Map<Player, Boolean> can_play; //If the red(blue) player can play or if it has no moves left
	Map<Player, int[]> penguins; //For every player, the list of his penguins
	Map<Player, Integer> score; //Score of every player
	Player humanPlayer;

	public GameState()
	{
		fish = new long[3];
		can_play = new HashMap<>();
		can_play.put(Player.Red, true);
		can_play.put(Player.Blue, true);
		penguins = new HashMap<>();
		penguins.put(Player.Red, new int[4]);
		penguins.put(Player.Blue, new int[4]);
		score = new HashMap<>();
		score.put(Player.Red, 0);
		score.put(Player.Blue, 0);
	}

	public void update(String jsonString)
	{
		//System.out.println(jsonString);
		JSONObject json = new JSONObject(jsonString);
		can_play.put(Player.Red, json.getJSONObject("can_play").getBoolean("red"));
		can_play.put(Player.Blue, json.getJSONObject("can_play").getBoolean("blue"));
		current_player = Player.valueOf(json.getString("current_player"));
		fish[0] = json.getJSONObject("bitboards").getLong("onefish");
		fish[1] = json.getJSONObject("bitboards").getLong("twofish");
		fish[2] = json.getJSONObject("bitboards").getLong("threefish");
		score.put(Player.Red, json.getJSONObject("score").getInt("red"));
		score.put(Player.Blue, json.getJSONObject("score").getInt("blue"));
		JSONArray jsonPeng = json.getJSONObject("penguins").getJSONArray("red");
		for (int i = 0; i < jsonPeng.length(); i++)
			penguins.get(Player.Red)[i] = jsonPeng.getInt(i);
		jsonPeng = json.getJSONObject("penguins").getJSONArray("blue");
		for (int i = 0; i < jsonPeng.length(); i++)
			penguins.get(Player.Blue)[i] = jsonPeng.getInt(i);
	}


	/**
	 * Set the amount of fish on one tile
	 *
	 * @param tileNb the tile on which we want to set the amount of fish
	 * @param fishNb The amount of fish (0-3)
	 */
	public void setFish(int tileNb, int fishNb)
	{
		for (int i = 0; i < 3; i++)
			if (i == fishNb - 1)
				fish[i] |= (long) 1 << tileNb;
			else
				fish[i] &= ~((long) 1 << tileNb);
	}

	/**
	 * Reset fish bitboards
	 */
	public void clearFish()
	{
		for (int i = 0; i < fish.length; i++)
			fish[i] = 0;
	}

	/**
	 * @param tileNb The number of the tile we want to know the value of
	 * @return The number of fish on the tile, or 0 if empty
	 */
	public int getNbFish(int tileNb)
	{
		if (((fish[0] >>> tileNb) & 1) == 1)
			return 1;
		else if (((fish[1] >>> tileNb) & 1) == 1)
			return 2;
		else if (((fish[2] >>> tileNb) & 1) == 1)
			return 3;
		return 0;
	}

	/**
	 * @param player The player for which we are searching the penguin position
	 * @param i      The number of the penguin for this player
	 * @return The position (0-59) of the penguin on the board
	 */
	public int getPenguinPos(Player player, int i)
	{ return (penguins.get(player)[i] & 63); }

	/**
	 * @param player The penguin's owner
	 * @param i      The number of the penguin for this player
	 * @return the total moves that this penguin can do
	 */
	public int getPenguinTotalMoves(Player player, int i)
	{ return ((penguins.get(player)[i] >>> 6) & 63); }

	/**
	 * Get maximum number of tiles that a penguin can do in one specific direction
	 */
	public int getNbMoves(Player player, int penguinNumber, Direction direction)
	{
		int shift = 0;
		switch (direction)
		{
			case A:
				shift = 12;
				break;
			case B:
				shift = 15;
				break;
			case C:
				shift = 18;
				break;
			case D:
				shift = 21;
				break;
			case E:
				shift = 24;
				break;
			case F:
				shift = 27;
				break;
		}
		return ((penguins.get(player)[penguinNumber] >>> shift) & 7);
	}

	/**
	 * "I want to move penguin 1 in direction B for 5 steps"
	 * "This is you move number 10 out of all your possible moves"
	 * Note: moves start at 0. Steps start at 1 for this function
	 */
	public int getPlayerMoveNumber(Player player, int penguinNumber, Direction direction, int nbSteps)
	{
		int[] peng = penguins.get(player);
		int total = 0;
		for (int i = 0; i < penguinNumber; i++) //We add the moves of all preceding penguins
			total += getPenguinTotalMoves(player, i);
		for (Direction d : Direction.values())
		{
			if (d.equals(direction))
				return total + nbSteps - 1;
			else
				total += getNbMoves(player, penguinNumber, d);
		}
		return -1;
	}

	/**
	 * @return The number of the penguin of this player that is on tile number tileNumber. Or -1 if there isn't any (note: there could be a penguin of the other player)
	 */
	public int getPenguinOnTile(Player player, int tileNumber)
	{
		for (int i = 0; i < penguins.get(player).length; i++)
			if (getPenguinPos(player, i) == tileNumber)
				return i;
		return -1;
	}

	public void setPenguin(Player player, int penguinNumber, int penguin)
	{
		penguins.get(player)[penguinNumber] = penguin;
	}

	/**
	 * @return The list of all tiles that could be reached by moving penguin penguinNb of player player
	 */
	public List<Integer> getReachableTiles(Player player, int penguinNb)
	{
		List<Integer> result = new ArrayList<>();
		int startingTile = getPenguinPos(player, penguinNb);
		for (int i = 1; i <= getNbMoves(player, penguinNb, Direction.A); i++)
			result.add(startingTile + i * 7);
		for (int i = 1; i <= getNbMoves(player, penguinNb, Direction.B); i++)
			result.add(startingTile - i);
		for (int i = 1; i <= getNbMoves(player, penguinNb, Direction.C); i++)
			result.add(startingTile - i * 8);
		for (int i = 1; i <= getNbMoves(player, penguinNb, Direction.D); i++)
			result.add(startingTile - i * 7);
		for (int i = 1; i <= getNbMoves(player, penguinNb, Direction.E); i++)
			result.add(startingTile + i);
		for (int i = 1; i <= getNbMoves(player, penguinNb, Direction.F); i++)
			result.add(startingTile + i * 8);

		return result;
	}

	/**
	 * @return The state as needed for initialisation of the game (just bitboards, penguins, and score)
	 */
	public String toGameInputJSON()
	{
		JSONObject result = new JSONObject();
		JSONObject jsonBitboards = new JSONObject();
		jsonBitboards.put("onefish", fish[0]);
		jsonBitboards.put("twofish", fish[1]);
		jsonBitboards.put("threefish", fish[2]);
		result.put("bitboards", jsonBitboards);
		result.put("current_player", Player.Red);
		JSONObject jsonPenguins = new JSONObject();
		jsonPenguins.put("red", penguins.get(Player.Red));
		jsonPenguins.put("blue", penguins.get(Player.Blue));
		result.put("penguins", jsonPenguins);
		JSONObject jsonScore = new JSONObject();
		jsonScore.put("red", score.get(Player.Red));
		jsonScore.put("blue", score.get(Player.Blue));
		result.put("score", jsonScore);

		return result.toString();
	}

	public boolean getCanPlay(Player player) { return can_play.get(player); }

	public Player getCurrent_player()
	{
		return current_player;
	}

	public int getScore(Player player)
	{
		return score.get(player);
	}

	public Player getHumanPlayer()
	{
		return humanPlayer;
	}

	public void setHumanPlayer(Player humanPlayer)
	{
		this.humanPlayer = humanPlayer;
	}

	public void setScore(Player player, int value)
	{
		score.put(player, value);
	}

	public enum Direction
	{
		A, B, C, D, E, F;
	}

}
